-- Time-stamp: <Sat Aug 16 1997 22:05:29 Stardate: [-31]9799.39 hwloidl>
--
-- Alpha-Beta Search
--
-- Board manipulation functions for tic-tac-toe
-- The board is represented as a list of rows with each row being a list.
-- There is no check for the correct size of the board in the type.
-- This version is parameterised over the board size (defined in Wins.hs)
--
-----------------------------------------------------------------------------

--@node Board Manipulation, ,Top, Top
--@chapter Board Manipulation

module Board where

--@menu
--* Imports::			
--* Types::			
--* Show Fcts::			
--* Update and Insert::		
--* Check::			
--* Inital Board::		
--* Instances::			
--@end menu

--@node Imports, Types, Board Manipulation, Board Manipulation
--@section Imports

--@include Wins.hs
import Wins

#if defined(STRATEGIES)
--@ifset strategies
--@include Strategies.lhs
--@end ifset
import Strategies
#else
--@ifclear strategies
--@include ParForce.hs
--@end ifclear
#endif

--@node Types, Show Fcts, Imports, Board Manipulation
--@section Types

type Board = [Row] 
type Row = [Piece]
data Piece = X | O | Empty deriving Eq
                                    {-!NFData!-}

--@node Show Fcts, Update and Insert, Types, Board Manipulation
--@section Show Fcts

showBoard :: Board -> String
showBoard rs = concat (zipWith (++) (map showRow rs) (repeat "\n------\n"))
               ++ "\n\n" 

showRow :: Row -> String
showRow ps {- | length ps /= board_size = error "showRow: illegal size"
           | otherwise -} = concat $ zipWith (++) (map showPiece ps) (repeat "|")

showPiece :: Piece -> String
showPiece X = "X"
showPiece O = "O"
showPiece Empty = " "

--@node Update and Insert, Check, Show Fcts, Board Manipulation
--@section Update and Insert

newPositions :: Piece -> Board -> [Board]
newPositions piece board = 
  _scc_ "newPositions" (  
    map (\ b -> if length b /= board_size
                                        then error ("newPositions: illegal board size " ++ (show (length b)) ++ (showBoard b))
					else b) $
				   -- concat {- parMap rnf -}
                                   map (placePiece piece board) 
				       [(x,y) | x<-[1..board_size],
				                y<-[1..board_size],
						empty (x,y) board]
						)

placePieces :: Board -> [(Piece, (Int,Int))] -> Board
placePieces = foldr (\ (p, pos) b -> placePiece p b pos) 

placePiece :: Piece -> Board -> (Int,Int) -> Board
--placePiece p board pos | not (empty pos board) = []
--placePiece p [r1,r2,r3] (1,x) = [[insert p r1 x,r2,r3]]
--placePiece p [r1,r2,r3] (2,x) = [[r1,insert p r2 x,r3]]
--placePiece p [r1,r2,r3] (3,x) = [[r1,r2,insert p r3 x]]
placePiece p b (x,y) = (take (x-1) b) ++          -- rows 1..x-1
                       [insert p (b!!(x-1)) y] ++ -- row  x
		       drop x b                   -- row x+1..board_size

insert :: Piece -> Row -> Int -> Row
--insert p [p1,p2,p3] 1 = [p,p2,p3]
--insert p [p1,p2,p3] 2 = [p1,p,p3]
--insert p [p1,p2,p3] 3 = [p1,p2,p]
insert p row n = take (n-1) row ++ [p] ++ drop n row

--@node Check, Inital Board, Update and Insert, Board Manipulation
--@section Check

empty :: (Int,Int) -> Board -> Bool
empty (x,y) b = (b!!(x-1))!!(y-1) == Empty

{- old (fixed board size)
empty :: (Int,Int) -> Board -> Bool
empty (1,x) [r1,r2,r3] = empty' x r1
empty (2,x) [r1,r2,r3] = empty' x r2
empty (3,x) [r1,r2,r3] = empty' x r3

empty' :: Int -> Row -> Bool
empty' 1 [Empty,_,_] = True
empty' 2 [_,Empty,_] = True
empty' 3 [_,_,Empty] = True
empty' _ _ = False
-}

fullBoard _ b = not (elem Empty (concat b))

--@node Inital Board, Instances, Check, Board Manipulation
--@section Inital Board

initialBoard :: Board
--initialBoard = [[Empty,Empty,Empty], 
--		[Empty,Empty,Empty],
--		[Empty,Empty,Empty]]
initialBoard = [ take board_size empties | i <- [1..board_size] ]
               where empties = repeat Empty
		     
#if (__HASKELL1__==4)
data Evaluation = XWin | OWin | Score Int deriving (Show,Eq)
                                                   {-!NFData!-}
#else
data Evaluation = XWin | OWin | Score Int deriving (Text,Eq)
                                                   {-!NFData!-}
#endif

{- OLD: partain
instance Eq Evaluation where
    XWin       == XWin	     = True
    OWin       == OWin	     = True
    (Score i1) == (Score i2) = i1 == i2
    _	       == _          = False
    a	  /= b	   = not (a == b)

instance Text Evaluation where
    showsPrec d XWin = showString "XWin"
    showsPrec d OWin = showString "OWin"
    showsPrec d (Score i) = showParen (d >= 10) showStr
	where
	  showStr = showString "Score" . showChar ' ' . showsPrec 10 i

    readsPrec p = error "no readsPrec for Evaluations"
    readList = error "no readList for Evaluations"
    showList []	= showString "[]"
    showList (x:xs)
		= showChar '[' . shows x . showl xs
		  where showl []     = showChar ']'
			showl (x:xs) = showChar ',' . shows x . showl xs
-}

instance Ord Evaluation where
 _ <= XWin    = True
 XWin <= _    = False
 OWin <= _    = True
 _ <= OWin    = False
 (Score x) <= (Score y) = x <= y

-- Positive values mean good position for X
-- Negative values mean good position for O
--eval 3 = XWin
--eval (-3) = OWin
--eval x = Score x
eval x | x >= board_size  = XWin
       | x <= -board_size = OWin
       | otherwise        = Score x
       
static :: Piece -> Board -> Evaluation
static _ board = interpret 0 (map (score board) wins)

--				`using` markStrat "static" (parList rnf))

{-
#if defined(PAR_STATIC) && defined(STRATEGIES)
static _ board = _scc_ "static"  ( interpret 0 (parMap rnf (score board) wins) )
#else
#endif
-}



interpret :: Int -> [Evaluation] -> Evaluation
interpret x [] = (Score x)
interpret x (Score y:l) = interpret (x+y) l
interpret x (XWin:l) = XWin
interpret x (OWin:l) = OWin

score :: Board -> Win -> Evaluation
score board win = eval (sum (map sum (zipWith (zipWith scorePiece) board win)))

scorePiece :: Piece -> Int -> Int
scorePiece X score = score
scorePiece Empty _ = 0
scorePiece O score = -score

{-
#if 0
-- This looks very much like a zipWith f to me
map2 :: (a -> b -> c) -> [a] -> [b] -> [c]
map2 f [] x = []
map2 f x [] = []
map2 f (x:xs) (y:ys) = f x y:map2 f xs ys
#endif
-}

--@node Instances,  , Inital Board, Board Manipulation
--@section Instances

{-* DO NOT EDIT BELOW THIS POINT *-}
instance NFData Piece where{rnf (X)= ();rnf (O)= ();rnf (Empty)= ()}
instance NFData Evaluation where{rnf (XWin)= ();rnf (OWin)= ();rnf (Score a)=rnf a}
