-- Time-stamp: <Sat Sep  6 1997 03:18:25 Stardate: [-31]9900.48 hwloidl>
--
-- Alpha-Beta Search
-- Search code i.e. fcts for determining the best next move.
-- Includes the top level pipeline in the program in `bestMove'
--
-----------------------------------------------------------------------------

--@node Making a Move, , ,
--@chapter Making a Move

module Game where

--@menu
--* Imports::			
--* Types::			
--* Main Fcts::			
--* Auxiliary Fcts::		
--* Min-Max Fcts::		
--* top level pipeline::	
--@end menu

--@node Imports, Types, Making a Move, Making a Move
--@section Imports

--@include Board.hs
--@include Wins.hs
--@include Tree.hs

import Board
import Wins
import Tree

#if defined(STRATEGIES)
-- --@include Strategies.lhs
import Strategies
#else
#endif

--@node Types, Main Fcts, Imports, Making a Move
--@section Types

#if defined(SIMPLE)
-- Either max or min on Evaluations
type Player = Evaluation -> Evaluation -> Evaluation
#else
type Player = [[Evaluation]] -> [Evaluation]
#endif
type Move = (Board,Evaluation)

--@node Main Fcts, Auxiliary Fcts, Types, Making a Move
--@section Main Fcts

-- How far to look ahead in the search tree (used to prune the tree)
--search_tree_depth :: Int
--search_tree_depth = 5

alternate :: Int -> Piece -> Player -> Player -> Board -> [Move]
alternate _ p _ _ b | fullBoard p b = []
alternate _ p _ _ b | static p b == XWin = []
alternate _ p _ _ b | static p b == OWin = []
alternate depth player f g board = move:alternate depth opposition g f board'
	where
	move@(board',eval) = best f possibles scores
	scores = 
#if defined(DATA_PAR) && defined(STRATEGIES)
	         parMap rnf 
#else
	         map 
#endif
	            (bestMove depth opposition g f) possibles
	possibles = newPositions player board
	opposition = opposite player

--@node Auxiliary Fcts, Min-Max Fcts, Main Fcts, Making a Move
--@section Auxiliary Fcts

opposite :: Piece -> Piece
opposite X = O
opposite O = X


best :: Player -> [Board] -> [Evaluation] -> Move
--best :: (Evaluation -> Evaluation -> Evaluation) -> 
--        [Board] -> [Evaluation] -> Move
#if defined(SIMPLE)
best f [] _ = error "No possible next move; prbly your fullBoard fct is buggy\n"
best f _ [] = error "No possible next move; prbly your fullBoard fct is buggy\n"
best f (b:bs) (s:ss) = best' b s bs ss
	where
	best' b s [] [] = (b,s)
	best' b s (b':bs) (s':ss) | s==(f s s') = best' b s bs ss
				  | otherwise 	= best' b' s' bs ss
#else
best f [] _ = error "No possible next move; prbly your fullBoard fct is buggy\n"
best f _ [] = error "No possible next move; prbly your fullBoard fct is buggy\n"
best f (b:bs) (s:ss) = best' b s bs ss
	where
	best' b s [] [] = (b,s)
	best' b s (b':bs) (s':ss) | s==(last (f [[s], [s']])) = best' b s bs ss
				  | otherwise 	       = best' b' s' bs ss
#endif

showMove :: Move -> String
showMove (b,e) = show e ++ "\n" ++ showBoard b

-- Cut-off all trees that statically evaluate to a winning position?? -- HWL
cropTree :: (Tree Evaluation) -> (Tree Evaluation)
cropTree (Branch a []) = (Branch a [])
cropTree (Branch (Score x) l) = Branch (Score x) 
	                         ({- parMap rnf -} map cropTree l)
cropTree (Branch x l) = Branch x []

{- 
   Inlined it into bestMove
searchTree :: Int -> Piece -> Board -> (Tree Board)
searchTree depth p board = prune depth           -- search_tree_depth
	              (repTree (newPositions p) (newPositions (opposite p)) board)
-}

#if defined(SIMPLE)
-- This does simple minimaxing without pruning subtrees based on intermediate
-- evaluations (i.e. purely compositional)
mise :: Int -> Int -> Player -> Player -> (Tree Evaluation) -> Evaluation
mise depth r f g (Branch a []) = _scc_ "mise" a
mise depth r f g (Branch _ l) = _scc_ "mise" ( foldr f (g OWin XWin) 
# if defined(DATA_PAR) && defined(STRATEGIES)
#  if (DATA_PAR_STRAT==0)
                         (map (mise depth (r-1) g f) l)
#  elif (DATA_PAR_STRAT==1)
                         (parMap rwhnf (mise depth (r-1) g f) l) -- see mmE-2.out, mm-4.out
#  elif (DATA_PAR_STRAT==2)
                         ((map (mise depth (r-1) g f) l) 
			  `using` (\ xs -> rnf (head xs) `par`
                                          parList rwhnf (tail xs) `par` ()))
#  elif (DATA_PAR_STRAT==3)
                         -- force the first n elems of the result list
                         ((map (mise depth (r-1) g f) l) 
			  `using` (\ _ -> if (FORCE_LEN)==(-1)  -- infinity 
			                  then parList rnf l `par` ()
			                  else parList rnf (take (FORCE_LEN) l) `par`
                                               parList rwhnf (drop (FORCE_LEN) l) `par` 
					       () 
					       ))
#  elif (DATA_PAR_STRAT==4)
                         -- force the first n/100 % of all elems
                         (
                         let
			   perc_step :: Int
			   perc_step = (FORCE_LEN_PERC_MAX-FORCE_LEN_PERC_MIN) `div` depth
                           fl :: Int			    
			   fl = (((FORCE_LEN_PERC_MIN) + perc_step * r) * (length l)) `div` 100
			 in
                         ((map (mise depth (r-1) g f) l) 
			  `using` (\ _ ->  parList rnf (take fl l) `par`
                                          parList rwhnf (drop fl l) `par` 
						 ()) ))
#  else
                         (parMap rwhnf (mise depth (r-1) g f) l)
#  endif /* DATA_PAR_STRAT */
# else  /* !DATA_PAR */
                         (map (mise depth (r-1) g f) l)
# endif
                        )
#else
-- A pruning version based on Hughes' paper
mise :: Int -> Int -> Player -> Player -> (Tree Evaluation) -> [Evaluation]
mise depth r f g (Branch a []) = [a]
mise depth r f g (Branch _ l) =  f 
# if defined(DATA_PAR) && defined(STRATEGIES)
#  if (DATA_PAR_STRAT==0)
                         (map (mise depth (r-1) g f) l)
#  elif (DATA_PAR_STRAT==1)
                         (parMap rwhnf (mise depth (r-1) g f) l) -- see mmE-2.out mm-4.out
#  elif (DATA_PAR_STRAT==2)
                         -- force the first elem of the result list
                         ((map (mise depth (r-1) g f) l)
			  `using` (\xs -> rnf (head xs) `par`
                                          parList rwhnf (tail xs) `par` ()))
#  elif (DATA_PAR_STRAT==3)
                         -- force the first n elems of the result list
                         ((map (mise depth (r-1) g f) l) 
			  `using` (\ _ -> if (FORCE_LEN)==(-1)  -- infinity 
			                  then parList rnf l `par` ()
			                  else parList rnf (take (FORCE_LEN) l) `par`
                                               parList rwhnf (drop (FORCE_LEN) l) `par` 
					       () 
					       ))
#  elif (DATA_PAR_STRAT==4)
                         -- force the first n/100 % of all elems
                         (
                         let
			   perc_step :: Int
			   perc_step = (FORCE_LEN_PERC_MAX-FORCE_LEN_PERC_MIN) `div` depth
                           fl :: Int			    
			   fl = (((FORCE_LEN_PERC_MIN) + perc_step * r) * (length l)) `div` 100
			 in
                         ((map (mise depth (r-1) g f) l) 
			  `using` (\ _ -> parList rnf   (take fl l)  `par`
                                         parList rwhnf (drop fl l)  `par` 
						 ()))
#  else
                         (parMap rwhnf (mise depth (r-1) g f) l)
#  endif /* DATA_PAR_STRAT */
# else  /* !DATA_PAR */
                           (map (mise depth (r-1) g f) l)
# endif
                           )
#endif

--@node Min-Max Fcts, top level pipeline, Auxiliary Fcts, Making a Move
--@section Min-Max Fcts

#if !defined(SIMPLE)
-- Input:   A list of decreasing lists (approximations to real value)
-- Output:  An increasing list (leading to the real value)
max' :: [[Evaluation]] -> [Evaluation]
max' (first:rest) = m : betterthan minleq minimum m rest
                    where m = minimum first  -- strict in first

min' :: [[Evaluation]] -> [Evaluation]
min' (first:rest) = m : betterthan maxleq maximum m rest
                    where m = maximum first  -- strict in first

-- Non strict over 3rd arg !!! [[Evaluation]]
-- Takes a list of approximations (list of Evals) and produces an approximation
-- by picking the worst value out of any approximation. Since the result list
-- must be increasing (wrt the `better' function) we can dump lists whose worst
-- value is worse than the current score determined (i.e. if we know that we 
-- can reach a position with value a and we know that the worst position coming
-- out of the move under consideration has a score less than that, then we do 
-- not have to look at all other positions in this subtree, because the we 
-- assume that the opponent will make the worst move for us).

betterthan :: (Evaluation -> [Evaluation] -> Bool) ->  -- maxleq or minleq 
              ([Evaluation] -> Evaluation) ->          -- max' or min'
	      Evaluation ->                            -- Score to compare it
	      [[Evaluation]] ->                        -- list of approxs 
	      [Evaluation]
betterthan _ _ _ [] = []
betterthan better_than_worst_of worst_of a (next:rest)
 | a `better_than_worst_of` next = betterthan better_than_worst_of worst_of a rest
 | otherwise                     = m : betterthan better_than_worst_of worst_of m rest
                                   where m = worst_of next

-- minleq y l <=> minimum l <= y
minleq :: Evaluation -> [Evaluation] -> Bool
minleq y []      = False
minleq y (x:xs)  
  | x <= y    = True       -- throws away the rest of the list!
  | otherwise = minleq y xs

-- maxleq l x <=> maximum l >= x
maxleq :: Evaluation -> [Evaluation] -> Bool
maxleq y []      = False
maxleq y (x:xs)  
  | x >= y    = True       -- throws away the rest of the list!
  | otherwise = maxleq y xs


#else
-- Evaluation is now an instance of Ord i.e. std min and max come for free
max' :: Evaluation -> Evaluation -> Evaluation
max' = max
min' :: Evaluation -> Evaluation -> Evaluation
min' = min

--max' :: Evaluation -> Evaluation -> Evaluation
--max' XWin _ = XWin
--max' _ XWin = XWin
--max' b OWin = b
--max' OWin b = b
--max' a@(Score x) b@(Score y) 	| x>y = a
--			 	| otherwise = b

--min' :: Evaluation -> Evaluation -> Evaluation
--min' OWin _ = OWin
--min' _ OWin = OWin
--min' b XWin = b
--min' XWin b = b
--min' a@(Score x) b@(Score y) 	| x<y = a
--				| otherwise = b
#endif
-- ---------------------------------------------------------------------------
--@node Top level pipeline,  , Min-Max Fcts, Making a Move
--@section Top level pipeline

bestMove :: Int -> Piece -> Player -> Player -> Board -> Evaluation
#if defined(SIMPLE)
bestMove depth p f g = (mise depth depth f g) . 
	               cropTree . 
		       (mapTree (static p)) . 
		       (prune depth) .
		       repTree (newPositions p) 
		               (newPositions (opposite p))
		       -- searchTree depth p
#elif defined(PAR_PIPE) && defined(STRICTIFY) && defined(STRATEGIES)
bestMove depth p f g = last               .|| rnf $
	               (mise depth depth f g)         .|| rnf $ 
	               cropTree           .|| rnf $ 
		       (mapTree (static p))   .|| rnf $ 
		       -- searchTree depth p
		       (prune depth)      .|| rwhnf $
	               repTree (newPositions p) 
		                (newPositions (opposite p))
#elif defined(PAR_PIPE) && defined(STRATEGIES)
bestMove depth p f g = last               .|| rwhnf $ -- list of approx
	               (mise depth depth f g)         .|| rwhnf $ -- cropped eval tree
	               cropTree           .|| rwhnf $ -- static eval tree
		       (mapTree (static p))   .|| rwhnf $ -- pruned search tree
		       -- searchTree depth p
		       (prune depth)      .|| rwhnf $ -- full search tree
	               repTree (newPositions p) 
		                (newPositions (opposite p))
#else
bestMove depth p f g = last .
	               (mise depth depth f g) . 
	               cropTree . 
		       (mapTree (static p)) . 
		       (prune depth) .
		       repTree (newPositions p) 
		               (newPositions (opposite p))
		       -- searchTree depth p
#endif

