-- Time-stamp: <Mon May 26 1997 23:35:29 Stardate: [-31]9389.70 hwloidl>
--
-- Alpha-Beta Search
-- Top level function
-- Takes an argument defining the depth in the search tree.
-- 
-- General setup:
--  SIMPLE ... don't prune the search tree based on intermediate evaluations;
--             this makes the algorithm compositional (using maximum to select 
--             best score), but does a lot of unnecessary work.
-- Kinds of parallelism:
--  PAR_PIPE ... parallel pipeline in getting best move
--  DATA_PAR ... data parallelism over all possible next moves
--  PAR_HO   ... parallel h.o. fcts (espec. over search trees)
--  PAR_STATIC ... parallel static evaluation function
--
-----------------------------------------------------------------------------

--@node Top  Level Fct, , Top, Top
--@chapter Top Level Fct

module Prog(prog) where

--@menu
--* Imports::			
--* Top Level Fct::		
--* Input Data::		
--@end menu

--@node Imports, Top Level Fct, Driver Fct, Driver Fct
--@section Imports

--@include Board.hs
--@include Wins.hs
--@include Game.hs
--@include Tree.hs

import Board
import Wins
import Game
import Tree
#if defined(STRATEGIES)
-- --@include Strategies.lhs
import Strategies
#endif

--@node Top Level Fct, Input Data, Imports, Driver Fct
--@section Top Level Fct

-- First arg is depth in the search tree
prog :: Int -> String -> String
prog depth _ = showMove (head game)
	       --"OXO\n" ++
	       --concat (map showMove game)
	       where
	       game = if depth <= 0 
	                then error "Search depth must be larger than 0\n"
#if (BOARD==1)
	                else alternate depth X max' min' initialBoard
#elif (BOARD==2)
	                else alternate depth O min' max' testBoard_II
#else			
	                else alternate depth X max' min' initialBoard
#endif

--@node Input Data,  , Top Level Fct, Driver Fct
--@section Input Data


-- O at center field; can escape in two moves
testBoard_II  = [Empty, Empty, Empty, Empty, Empty, O, X, X, X, Empty, Empty]
--testBoard_III = [[Empty,Empty,Empty],[Empty,X,Empty],[Empty,Empty,Empty]]
