-- Time-stamp: <Wed Jun 18 1997 20:35:38 Stardate: [-31]9504.08 hwloidl>
--
-- Alpha-Beta Search
-- Top level function
-- Takes an argument defining the depth in the search tree.
-- 
-- General setup:
--  SIMPLE ... don't prune the search tree based on intermediate evaluations;
--             this makes the algorithm compositional (using maximum to select 
--             best score), but does a lot of unnecessary work.
-- Kinds of parallelism:
--  PAR_PIPE ... parallel pipeline in getting best move
--  DATA_PAR ... data parallelism over all possible next moves
--  PAR_HO   ... parallel h.o. fcts (espec. over search trees)
--  PAR_STATIC ... parallel static evaluation function
--
-----------------------------------------------------------------------------

--@node Top  Level Fct, , Top, Top
--@chapter Top Level Fct

module Prog(prog) where

--@menu
--* Imports::			
--* Top Level Fct::		
--* Input Data::		
--@end menu

--@node Imports, Top Level Fct, Driver Fct, Driver Fct
--@section Imports

--@include Board.hs
--@include Wins.hs
--@include Game.hs
--@include Tree.hs

import Board
import Wins
import Game
import Tree
#if defined(STRATEGIES)
-- --@include Strategies.lhs
import Strategies
#endif

--@node Top Level Fct, Input Data, Imports, Driver Fct
--@section Top Level Fct

-- First arg is depth in the search tree
prog :: Int -> String -> String
prog depth _ = showMove (head game)
	       --"OXO\n" ++
	       -- concat (map showMove game)
	       where
	       game = if depth <= 0 
	                then error "Search depth must be larger than 0\n"
#if (BOARD==1)
			else alternate depth X max' min' testBoard_I
#elif (BOARD==2)
			else alternate depth X max' min' testBoard_II
#elif (BOARD==3)
			else alternate depth O min' max' testBoard_III
#else
			else alternate depth X max' min' testBoard_I
#endif

--@node Input Data,  , Top Level Fct, Driver Fct
--@section Input Data

testBoard_I   = placePieces initialBoard  [(X, (2,2)), (O, (1,2))] 
testBoard_II  = placePieces initialBoard  [(X, (2,2)), (O, (1,2)), (X, (1,1)), (O, (3,3))] 
testBoard_III = placePieces initialBoard  [(X, (2,2))]

--testBoard_I   = [[Empty,O,Empty],[Empty,X,Empty],[Empty,Empty,Empty]]
--testBoard_II  = [[X,O,Empty],[Empty,X,Empty],[Empty,Empty,O]]
--testBoard_III = [[Empty,Empty,Empty],[Empty,X,Empty],[Empty,Empty,Empty]]
